#!/usr/bin/perl

# Converts fractint sytle lsystem files to L-system Explorer's .lsysexp format
# 
# Copyright (c) 2002, Ron Coscorrosa.  Released under the terms of the GNU GPL
# 
# Usage:  perl fractint2lsysexp.pl file1 file2 file3 >> ~/.lsysexp
#
# If the first argument is "include" it will output a C header (.h) file that
# you can use to compile into L-system Explorer, i.e.:
#
#   perl fractint2lsysexp.pl include file1 file2 > lsysexp/src/builtin.h

use strict;

my $total;
my %fractals;
my $mode = ($ARGV[0] eq "include") ? shift(@ARGV) : "rc";

my %default = ('name' => undef,
	       'rules' => undef,
	       'axiom' => undef,
	       'da' => 12,
	       'a0' => 90.0);

scan_lsystems();

if ($mode eq "rc")
{
   output_as_rc();
}
else
{
   output_as_include();
}
print STDERR "$total L-systems converted\n";

sub scan_lsystems
{
  for (@ARGV)
  {
    open (INPUT, "$_") or next;
    my $group = $_;
    $group =~ s/^.*[\\\/]//;
    $group =~ s/\.l$//i;
    $group =~ s/\W//g;
    $default{group} = $group;
    my %lsys = %default;
    while (<INPUT>)
      {
	chomp;
	s/\;.*$//;       # strip comments, leading and trailing whitespace
	s/^\s*//;
	s/\s*$//;
	next if /^\s*$/;
	
	# Found }, output L-system
	if (/^}$/)
	{
	  if ($lsys{name} && $lsys{rules} && $lsys{axiom})
	    {
	      push (@{$fractals{$group}}, {%lsys});
	    }
	  %lsys = %default;
	}
        elsif (/^\s*Angle\s*=?\s*(\d+)/i)
	{
	  $lsys{da} = 360.0 / $1;
	}
        elsif (/^\s*Axiom\s*([\(\)\\\/.\w\+\-!@|\/\[\]><]+)/i)
	{
	  $lsys{axiom} = convert($1);
	}
        elsif (/\s*(\w)=([\)\(\\\/.\w\+\-!@|\/\[\]<>]+)?.*$/)
	{
	  $lsys{rules}->{$1} .= convert($2);
	}
        elsif (/^([\w\d\-\&]+)\s*{?/)
	{
	  $lsys{name} = $1;
	  $lsys{name} =~ s/\&//g;
	  $lsys{name} =~ s/\-/_/g;
	}
	elsif ( !/^{$/)
	{
	  print STDERR "Ignoring \"$_\"\n";
	}
    }
    close (INPUT);
  }
}


# Output a .lsystemrc style thing
sub output_as_rc
{
  for (sort {$a cmp $b} keys (%fractals))
  {
    for (@{$fractals{$_}})
      {
	my $lsys = $_;
	my $rules = $lsys->{rules};
	my $bleh = join ("\n", map {"rules: $_ = $rules->{$_}" }keys (%{$rules}));
	++$total;
	print "{\n";
	print "group: $lsys->{group}\n";
	print " name: $lsys->{name}\n";
	print "axiom: $lsys->{axiom}\n";
	print "   da: $lsys->{da}\n";
	print "   a0: $lsys->{a0}\n";
	print "$bleh\n";
	print "}\n";
      }
  }
}



# Output in a format suitable to be included as "builtin.h"
sub output_as_include
{
  my ($i, $last, @keys);
  @keys = sort {$a cmp $b} keys %fractals;

  print <<__INCLUDE_HEADER__;
#ifndef BUILTIN_H
#define BUILTIN_H

#include "util.h"
#include "constants.h"
__INCLUDE_HEADER__

  for (@keys)
    {
      $last = scalar(@{$fractals{$_}}) - 1;
      print "static lsys $_ [] = \n   {\n";
      for $i(0 .. @{$fractals{$_}} - 1)
	{
	  my $lsys = $fractals{$_}->[$i];
	  my $rules = $lsys->{rules};
	  my $r = join (";", map {"$_ = $rules->{$_}" }keys (%{$rules}));
	  my $comma = ($i eq $last) ? "\n};\n" : ",";
           
	  $total++;

	  $lsys->{axiom} =~ s#\\#\\\\#g;
	  $r =~ s#\\#\\\\#g;
	  print <<__END_OF_LSYS__;
    {2, $lsys->{da}, $lsys->{a0}, .50, "$lsys->{axiom}", "$r", "$lsys->{name}"}$comma
__END_OF_LSYS__
	}
    }

  print <<__STRUCT__;
static struct
{
  char * group_name;
  lsys * l;
  int size;
} builtins [] =
  {
__STRUCT__

   for $i (0 .. $#keys)
   {
     print "     {\"$keys[$i]\", $keys[$i], countof($keys[$i])}";
     print "," if ($i != $#keys);
     print "\n";
   }
print <<__END_OF_INCLUDE__;
};
#endif /* BUILTIN_H */
__END_OF_INCLUDE__
}



# Convert fractint style formulas:
# 1) Switch '+' and '-'
# 2) Get rid of all the color commands:  Cxx, <xx, >xx
sub convert
{
  my $orig = shift;
  my $new;
  $orig =~ s#c\d+##gi;
  $orig =~ s#<\d+##gi;
  $orig =~ s#>\d+##gi;

  for (split ('', $orig))
    {
      # swap + and -
      $new .= ($_ eq "+") ? "-" : ($_ eq "-") ? "+" : $_;
    }
  return $new;
}

